require 'soar/registry/directory/error'
require 'hashie'
require 'mince'
require 'hashy_db'

module Soar
  module Registry
    module Directory
      module Provider
        class Stub

          ##
          # @param [String] table the table name
          # @param [Array<String>] indexes primary key should be the first item 
          ##
          def initialize(table: , index: )
            @interface = Mince::HashyDb::Interface
            @table = table
            @primary_key = index[0]
            @index = index
            @interface.clear
          end

          ##
          # @param [Hash{String => Hash, String, Number, Array}] entry
          # @raise [Soar::Registry::Directory::Error::DuplicateEntryError]
          # @return [Boolean]
          ##
          def put(entry)
            adapt_exceptions do
              if not @interface.get_for_key_with_value(@table, @primary_key, entry[@primary_key])
                 @interface.add(@table, entry)
                 return true
              end
              raise Soar::Registry::Directory::Error::DuplicateEntryError, "Unable to add duplicate entry for index #{@primary_key} with value #{entry[@primary_key]}"
            end
          end

          ##
          # @param [String] primary_key
          # @return [Hash{String => Hash, String, Number, Array}] entry
          # @raise [Soar::Registry:::Directory::Error::NoEntriesFoundError] if primary key not found
          ##
          def fetch(primary_key)
            adapt_exceptions do
              entry = @interface.get_for_key_with_value(@table, @primary_key, primary_key)
              raise Soar::Registry::Directory::Error::NoEntriesFoundError, "No entries found for #{@primary_key} = #{primary_key}" if entry.nil?
              entry
            end
          end

          ##
          # @param [String] key
          # @param [String] value 
          # @return [Array<String>] list of matching entries
          # @raise [ArgumentError] if query or index is not specified
          ##
          def search(key, value)
            adapt_exceptions do
              identities = []
              @interface.find_all(@table).each { |identity| 
                if identity[key] == value
                  identities << identity
                else
                  next
                end
              }
              identities
            end
          end

          ##
          # @return [Array<String>] a list of provider specific indexes
          ##
          def index
            adapt_exceptions do
              [@interface.primary_key.to_s]
              @index
            end
          end


          private 

          def adapt_exceptions
            begin
              yield
            rescue NoMethodError => e
              raise Soar::Registry::Directory::Error::NetworkingError, 'Networking error'
            end
          end

        end

      end
    end
  end
end
