require 'soar/registry/directory'
require 'securerandom'

module Soar
  module Registry
    module Directory
      module Test
        module Provider
          class Ldap

            def initialize
              @configuration = YAML.load_file("config/#{ENV['CONFIG_FILE']}")
              @base = @configuration['provider']['config']['base']
              @index = @configuration['provider']['config']['index']
              @entries = [{
                "dn" => "cn=John Smith,#{@base}",
                "attributes" => {
                  "cn" => "John Smith",
                  "mail" => "test@example.com",
                  "objectclass" => ["inetOrgPerson", "top"],
                  "sn" => "Smith"
                }
              }, {
                "dn" => "cn=Sarel Cilliers,#{@base}",
                "attributes" => {
                  "cn" => "Sarel Cilliers",
                  "mail" => "sarel@example.com",
                  "objectclass" => ["inetOrgPerson", "top"],
                  "sn" => "Cilliers"
                }
              }]
              @connection_retries = 6
            end

            def given_configured_directory
              begin
                provider_class = Object.const_get(@configuration['provider']['class'])
                provider = provider_class.new(@configuration['provider']['config'].map { |k,v| [k.to_sym, v] }.to_h )
                @directory = Soar::Registry::Directory.new(provider)
                @entries.each { |entry|
                  provider.delete(entry['dn'])
                }
              rescue Soar::Registry::Directory::Error::NetworkingError => e
                raise if @connection_retries == 0
                sleep 10
                @connection_retries -= 1
                given_configured_directory
              end
            end

            ##
            # Adds @entries to datastore using ldap protocol
            # Adds entryuuid to @entries
            ##
            def given_existing_data
              @entries.each{ |entry| 
                @directory.put(entry)
                result = @directory.search("mail", entry['attributes']['mail'])
                entry['attributes']['entryuuid'] = result[0][@index[0]]
              }
            end

            def sabotage_network
              @directory.provider.client = Net::LDAP.new({
                host: 'localhost',
                port: @configuration['provider']['config']['config']['port'],
                auth: {
                  method: :simple,
                  username: @configuration['provider']['config']['credentials']['username'],
                  password: @configuration['provider']['config']['credentials']['password']
                }              
              })
            end

            def put_duplicate_entry
              begin
                @directory.put(@entries[0])
              rescue Soar::Registry::Directory::Error::DuplicateEntryError => e
                @error = e
              end
            end

            def put_entry
              begin
                @directory.put(@entries[0])
                result = @directory.search("mail", @entries[0]['attributes']['mail'])
                @entries[0]['attributes']['entryuuid'] = result[0][@index[0]]
              rescue Soar::Registry::Directory::Error::NetworkingError => e
                @error = e
              end
            end

            def search_for_entry
              begin
                @result = @directory.search("mail", @entries[0]['attributes']['mail'])
              rescue Soar::Registry::Directory::Error::NetworkingError => e
                @error = e
              end
            end

            def request_index
              begin
                @result = @directory.index
              rescue Soar::Registry::Directory::Error::NetworkingError => e
                @error = e
              end
            end

            def fetch_entry
              begin
                @result = @directory.fetch(@entries[0]['attributes']['entryuuid'])
              rescue Soar::Registry::Directory::Error::NoEntriesFoundError, Soar::Registry::Directory::Error::NetworkingError => e
                @error = e
              end
            end

            def duplicate_entry?
              @error.is_a?(Soar::Registry::Directory::Error::DuplicateEntryError)
            end

            def persisted?
              result = @directory.fetch(@entries[0]['attributes']['entryuuid'])
              entry = @entries[0]['attributes']
              entry.delete('objectclass')
              result == entry
            end

            def returned?
              entry = @entries[0]['attributes']
              entry.delete('objectclass')
              @result == [entry]
            end

            def index?
              @index.is_a?(Array)
            end

            def single_entry?
              entry = @entries[0]['attributes']
              entry.delete('objectclass')
              @result == entry
            end

            def no_entries_found?
              @error.is_a?(Soar::Registry::Directory::Error::NoEntriesFoundError)
            end

            def no_matching_entries?
              @result == []
            end

            def networking_error?
              @error.is_a?(Soar::Registry::Directory::Error::NetworkingError)
            end

          end
        end
      end
    end

  end
end
